/*
 *  Copyright 2020-2023 Google LLC
 *  Copyright 2020-2023 EPAM Systems, Inc
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.opengroup.osdu.notification.provider.gcp.service;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNull;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

import java.util.Collections;
import java.util.HashSet;
import java.util.Set;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.junit.MockitoJUnitRunner;
import org.opengroup.osdu.core.common.model.notification.Subscription;
import org.opengroup.osdu.notification.provider.gcp.repo.SubscriptionCacheRepo;

@RunWith(MockitoJUnitRunner.class)
public class ExternalSubscriptionsManagerTest {

  public static final String TEST_TENANT = "test-tenant";
  private static final String TENANT_WO_SUBS = "no-sub-tenant";
  public static final String TEST_TOPIC = "topic";
  public static final String TEST_NOTIF_ID = "test-id";

  @InjectMocks
  private ExternalSubscriptionsManager manager;
  @Mock
  private SubscriptionCacheRepo mockSubCacheRepo;
  @Mock
  private RegisterSubscriptionService mockSubService;

  private Subscription subscription;

  private Set<Subscription> subscriptions;

  @Before
  public void setUp() {
    subscription = new Subscription();
    subscription.setNotificationId(TEST_NOTIF_ID);
    subscription.setTopic(TEST_TOPIC);

    subscriptions = new HashSet<>() {{
      add(subscription);
    }};
  }

  @Test
  public void testInvokeTenantSubscribers_noActiveSubs() {
    when(mockSubCacheRepo.getDataPartitionSubscribers(TENANT_WO_SUBS)).thenReturn(
        Collections.emptySet());
    manager.invokeTenantSubscribers(TENANT_WO_SUBS);
    verify(mockSubCacheRepo, times(0)).writeSubscription(anyString(), any());
    verify(mockSubService, times(1)).getAllSubscriptionInfos(TENANT_WO_SUBS);
  }

  @Test
  public void testInvokeTenantSubscribers_withActiveSubs() {
    when(mockSubCacheRepo.getDataPartitionSubscribers(TEST_TENANT)).thenReturn(subscriptions);
    manager.invokeTenantSubscribers(TEST_TENANT);
    verify(mockSubService, times(0)).getAllSubscriptionInfos(TEST_TENANT);
  }

  @Test
  public void testInvokeTenantSubscribers_withActiveSubs_emptyCache() {
    when(mockSubCacheRepo.getDataPartitionSubscribers(TEST_TENANT)).thenReturn(null);
    when(mockSubService.getAllSubscriptionInfos(TEST_TENANT)).thenReturn(Collections.singletonList(subscription));
    when(mockSubService.getSubscriptionsByNotificationId(TEST_TENANT, TEST_NOTIF_ID)).thenReturn(subscription);

    manager.invokeTenantSubscribers(TEST_TENANT);

    verify(mockSubService, times(1)).getAllSubscriptionInfos(TEST_TENANT);
    verify(mockSubCacheRepo, times(1)).writeSubscription(TEST_TENANT, subscription);
  }

  @Test
  public void testGetSubscriptionsForTopic() {
    doReturn(subscriptions).when(mockSubCacheRepo).getTopicSubscribers(TEST_TENANT, TEST_TOPIC);
    Set<Subscription> actualSubscriptions = manager.getSubscriptionsForTopic(TEST_TENANT,
        TEST_TOPIC);
    assertEquals(subscriptions, actualSubscriptions);
  }

  @Test
  public void testGetSubscription_inCache() {
    when(mockSubCacheRepo.readSubscription(TEST_TENANT, TEST_TOPIC, TEST_NOTIF_ID)).thenReturn(
        subscription);
    Subscription actualSub = manager.getSubscription(TEST_TENANT, TEST_TOPIC, TEST_NOTIF_ID);
    assertEquals(subscription, actualSub);
    verify(mockSubService, times(0)).getSubscriptionsByNotificationId(anyString(), anyString());
  }

  @Test
  public void testGetSubscription_notInCache_active() {
    when(mockSubCacheRepo.readSubscription(TEST_TENANT, TEST_TOPIC, TEST_NOTIF_ID)).thenReturn(null);
    when(mockSubService.getSubscriptionsByNotificationId(TEST_TENANT, TEST_NOTIF_ID)).thenReturn(
        subscription);

    Subscription actualSub = manager.getSubscription(TEST_TENANT, TEST_TOPIC, TEST_NOTIF_ID);
    assertEquals(subscription, actualSub);
    verify(mockSubCacheRepo, times(1)).writeSubscription(TEST_TENANT, subscription);
  }

  @Test
  public void testGetSubscription_notInCache_notAvailableInRegister() {
    when(mockSubCacheRepo.readSubscription(TEST_TENANT, TEST_TOPIC, TEST_NOTIF_ID)).thenReturn(null);
    when(mockSubService.getSubscriptionsByNotificationId(TEST_TENANT, TEST_NOTIF_ID)).thenReturn(
        null);

    Subscription actualSub = manager.getSubscription(TEST_TENANT, TEST_TOPIC, TEST_NOTIF_ID);
    assertNull(actualSub);
    verify(mockSubCacheRepo, times(1)).poisonSub(TEST_TENANT, TEST_TOPIC, TEST_NOTIF_ID);
  }

  @Test
  public void testGetSubscription_InCache_Poisoned() {
    when(mockSubCacheRepo.readSubscription(TEST_TENANT, TEST_TOPIC, TEST_NOTIF_ID)).thenReturn(null);
    when(mockSubCacheRepo.checkIfSubIsPoisoned(TEST_TENANT, TEST_TOPIC, TEST_NOTIF_ID)).thenReturn(
        new Subscription());

    Subscription actualSub = manager.getSubscription(TEST_TENANT, TEST_TOPIC, TEST_NOTIF_ID);
    assertNull(actualSub);
    verify(mockSubService, times(0)).getSubscriptionsByNotificationId(anyString(), anyString());
  }

  @Test
  public void testRemoveSubscription() {
    boolean b = manager.removeSubscription(TEST_TENANT, TEST_TOPIC, TEST_NOTIF_ID);
    verify(mockSubCacheRepo, times(1)).deleteSubscription(TEST_TENANT, TEST_TOPIC, TEST_NOTIF_ID);
  }

  @Test
  public void testUpdateSubscription_availableInRegister() {
    when(mockSubService.getSubscriptionsByNotificationId(TEST_TENANT, TEST_NOTIF_ID)).thenReturn(
        subscription);
    manager.updateSubscription(TEST_TENANT, TEST_NOTIF_ID);
    verify(mockSubCacheRepo, times(1)).writeSubscription(TEST_TENANT, subscription);
  }

  @Test
  public void testUpdateSubscription_notAvailableInRegister() {
    manager.updateSubscription(TEST_TENANT, TEST_NOTIF_ID);
    verify(mockSubCacheRepo, times(0)).writeSubscription(anyString(), any());
  }
}